# (C) Copyright 2019-2021 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.interface import Interface as AbstractInterface


class Interface(AbstractInterface):
    """
    Provide configuration management for Interface for Version 10.08.
    """

    pass
